import Instance.IndexedInstance;
import Propagators.ArcConsistencyPropagator;
import Propagators.CycleConsistencyPropagator;
import Propagators.PropagatorInterface;
import Propagators.SingletonArcConsistencyPropagator;
import Utils.*;

import java.util.LinkedList;

public class Optimizer {

    private Instance.IndexedInstance indexedInstance;
    private SparseVector h;
    private SetOfIndices S;
    private double[] weights;
    private SequenceSh sequenceSh;
    private boolean[] csp;
    private boolean[] cspCopy;
    private SetsUSf setsU;

    private int epsilonDecreaseRepetitions = 20;

    public Optimizer(Instance.IndexedInstance indexedInstance){
        this.indexedInstance = indexedInstance;
        h = new SparseVector(indexedInstance.tuples()+1);
        S = new SetOfIndices(indexedInstance.tuples()+1);
        weights = new double[indexedInstance.tuples()+1];
        sequenceSh = new SequenceSh(indexedInstance.tuples()+1);
        csp = new boolean[indexedInstance.tuples()+1];
        cspCopy = new boolean[indexedInstance.tuples()+1];
        setsU = new SetsUSf(indexedInstance.scopes());
    }

    public double getBound(PropagatorInterface propagator, long timeLimit){

        System.arraycopy(indexedInstance.tupleWeight, 0, weights, 0, indexedInstance.tupleWeight.length); // copy to weights
        double epsilon = calculateEpsilon(weights);
        double prevBound=Double.MAX_VALUE;
        double bound = calculateBound(weights,indexedInstance);
        boolean wasInconsistent = false;
        int remainingTrials = epsilonDecreaseRepetitions;
        while(true){
            System.out.println("Obj: " + bound + "       Eps: "+epsilon);
            if(System.nanoTime()>= timeLimit){
                return bound;
            }
            calculateActiveTuples(weights,epsilon,csp);
            System.arraycopy(csp, 0, cspCopy, 0, csp.length); // copy CSP to allow changing in propagator
            if(propagator.run(cspCopy, sequenceSh, setsU) == 1){
                if(epsilon < 1e-6){
                    break;
                } else {
                    if(!wasInconsistent){
                        epsilon = epsilon/10; // did not improve from last time -> decrease epsilon more significantly
                    } else {
                        epsilon = epsilon/10; // did improve from last time -> decrease only slightly
                    }
                    remainingTrials = epsilonDecreaseRepetitions;

                    wasInconsistent = false;
                    //System.out.println("Epsilon: " + epsilon);
                    continue;
                }
            }
            wasInconsistent = true;
            if((prevBound-bound)/Math.abs(prevBound) < 1e-15 || prevBound == bound){
                remainingTrials--;
                if(remainingTrials <= 0){
                    epsilon = epsilon/10;
                    if(epsilon < 1e-6){
                        break;
                    }
                    remainingTrials = epsilonDecreaseRepetitions;
                }
            } else {
                remainingTrials = epsilonDecreaseRepetitions;
            }
            // AC closure empty here
            constructSparseVectorHFromSequence(sequenceSh,setsU, h, S);
            double stepSize = Math.min(calculateBeta(weights,h),calculateGamma(weights,csp,setsU,h));
            //System.out.println("   step size: " + stepSize + " setsUlength: " + setsU.numElements());
            // perform step in weights
            for (int i = 0; i < h.numberOfNonzeroElements(); i++) {
                int tupleId = h.getIndexOfNthNonzeroElement(i);
                weights[tupleId] += stepSize*h.getValue(tupleId);
            }
            prevBound = bound;
            bound = calculateBound(weights, indexedInstance);
            if(System.nanoTime()>= timeLimit){
                return bound;
            }
        }


        return bound;
    }

    public double calculateTrueOptVal(double[] weights){
        // optimal value

        int[] labelling = new int[indexedInstance.nodes()+1];
        double maxValue = -Double.MAX_VALUE;
        while(true) {
            double currValue = 0;

            for (int edgeId = 1; edgeId <= indexedInstance.edges(); edgeId++) {
                currValue += weights[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId, labelling[indexedInstance.edgeIdFrom(edgeId)],
                        labelling[indexedInstance.edgeIdTo(edgeId)])];
            }

            for (int nodeId = 1; nodeId <= indexedInstance.nodes(); nodeId++) {
                currValue += weights[indexedInstance.nodeIdLabelToTupleId(nodeId, labelling[nodeId])];
            }

            maxValue = Math.max(currValue, maxValue);

            // pass to next labelling:
            int i;
            for (i = 1; i <= indexedInstance.nodes() + 1; i++) {
                if (i == indexedInstance.nodes() + 1) {
                    break;
                }
                if (labelling[i] + 1 < indexedInstance.domainSize(i)) {
                    break;
                }
            }
            if (i == indexedInstance.nodes() + 1) {
                break;
            } else {
                for (int j = 0; j < i; j++) {
                    labelling[j] = 0;
                }
                labelling[i]++;
            }
        }
        return maxValue;

    }

    private double calculateBeta(double[] weights, SparseVector h){
        double beta = Double.MAX_VALUE;
        for (int i = 0; i < h.numberOfNonzeroElements(); i++) {
            int tupleId = h.getIndexOfNthNonzeroElement(i);
            if(h.getValue(tupleId) > 0){
                // calculate max value in group of tuples (within the specified scope)
                double maxValueInGroupOfTuples = -Double.MAX_VALUE;
                int scopeId = indexedInstance.tupleIdToScopeId(tupleId);
                if(indexedInstance.arity(scopeId) == 1){
                    // node
                    int nodeId = indexedInstance.scopeIdToNodeId(scopeId);
                    for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                        maxValueInGroupOfTuples = Math.max(maxValueInGroupOfTuples,
                                weights[indexedInstance.nodeIdLabelToTupleId(nodeId,label)]);
                    }
                } else {
                    // edge
                    int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
                    for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); label1++) {
                        for (int label2 = 0; label2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); label2++) {
                            maxValueInGroupOfTuples = Math.max(maxValueInGroupOfTuples,
                                    weights[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2)]);
                        }
                    }
                }
                beta = Math.min(beta, (maxValueInGroupOfTuples-weights[tupleId])/h.getValue(tupleId));
            }
        }
        return beta;
    }

    private double calculateGamma(double[] weights, boolean[] csp, SetsUSf setsU, SparseVector h){
        // TODO: Maybe avoid copying CSP but decide activity based on weights
        double gamma = Double.MAX_VALUE;
        for (int i = 0; i < setsU.numElements(); i++) {
            int scopeId = setsU.getElement(i);
            if(indexedInstance.arity(scopeId) == 1){
                // node
                int nodeId = indexedInstance.scopeIdToNodeId(scopeId);
                for (int labelInactive = 0; labelInactive < indexedInstance.domainSize(nodeId); labelInactive++) {
                    int tupleIdInactive = indexedInstance.nodeIdLabelToTupleId(nodeId,labelInactive);
                    if(!csp[tupleIdInactive]){
                        for (int labelActive = 0; labelActive < indexedInstance.domainSize(nodeId); labelActive++) {
                            int tupleIdActive = indexedInstance.nodeIdLabelToTupleId(nodeId,labelActive);
                            if(csp[tupleIdActive] && h.getValue(tupleIdInactive) > h.getValue(tupleIdActive)){
                                gamma = Math.min(gamma,
                                        (weights[tupleIdActive]-weights[tupleIdInactive])/(h.getValue(tupleIdInactive)-h.getValue(tupleIdActive)));
                            }
                        }
                    }
                }
            } else {
                // edge
                int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
                for (int labelInactive1 = 0; labelInactive1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); labelInactive1++) {
                    for (int labelInactive2 = 0; labelInactive2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); labelInactive2++) {
                        int tupleIdInactive = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,labelInactive1,labelInactive2);
                        if(!csp[tupleIdInactive]){
                            for (int labelActive1 = 0; labelActive1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); labelActive1++) {
                                for (int labelActive2 = 0; labelActive2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); labelActive2++) {
                                    int tupleIdActive = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,labelActive1,labelActive2);
                                    if(csp[tupleIdActive] && h.getValue(tupleIdInactive) > h.getValue(tupleIdActive)){
                                        gamma = Math.min(gamma,
                                                (weights[tupleIdActive]-weights[tupleIdInactive])/(h.getValue(tupleIdInactive)-h.getValue(tupleIdActive)));
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return gamma;
    }

    private void constructSparseVectorHFromSequence(SequenceSh Sh, SetsUSf setsUSf, SparseVector h, SetOfIndices S){
        h.erase();
        S.erase();

        int initialSeqElem = -1;
        //System.out.println("Length of seq: " + Sh.getLengthOfSequence());
        for (int seqElem = Sh.getLengthOfSequence()-1; seqElem >= 0 ; seqElem--) {
            /*System.out.println(" Calc initSEQ: "+ seqElem);
            System.out.println(" Before processing: " + Sh.getSizeOfSetS(seqElem));*/
            // does setsUSf contain at least one scope from the specified set S?
            for (int i = 0; i < Sh.getSizeOfSetS(seqElem); i++) {
                if(setsUSf.contains(indexedInstance.tupleIdToScopeId(Sh.getElementOfSetS(seqElem,i)))){
                    initialSeqElem = seqElem;
                    break;
                }
                /*System.out.println(" Calc initseq: "+ seqElem + " tupleid: " + Sh.getElementOfSetS(seqElem,i) + " (scope: " + indexedInstance
                        .tupleIdToScopeId(Sh
                        .getElementOfSetS(seqElem,i)) + ")  is not in USf.");*/
            }
            if(initialSeqElem!=-1){
                break;
            }
        }
        if(initialSeqElem == -1){
            System.out.println("ERR: Not possible to construct deactivating vector");
        }
        //System.out.println("Initial seq elem: " + initialSeqElem);

        // initialize h as initialSeqElem-th vector h in Sh
        for (int i = 0; i < Sh.getNumberOfNonzeroElementsOfVecH(initialSeqElem); i++) {
            h.addValue(Sh.getNonZeroTupleIdOfVecH(initialSeqElem,i),Sh.getNonZeroValueOfVecH(initialSeqElem,i));
        }
        // add S_r to S
        for (int i = 0; i < Sh.getSizeOfSetS(initialSeqElem); i++) {
            S.add(Sh.getElementOfSetS(initialSeqElem,i));
            //System.out.println("Adding to S: " + Sh.getElementOfSetS(initialSeqElem,i));
        }

        // iterate over other h vectors in Sh downwards
        for (int seqElem = initialSeqElem-1; seqElem >= 0 ; seqElem--) {
            boolean addToH = false;
            for (int i = 0; i < Sh.getSizeOfSetS(seqElem); i++) {
                if(setsUSf.contains(indexedInstance.tupleIdToScopeId(Sh.getElementOfSetS(seqElem,i)))){
                    addToH = true;
                    break;
                }
            }
            if(!addToH){
                // iterate through set S and see if h_t = 0 for all t in S
                for (int i = 0; i < Sh.getSizeOfSetS(seqElem); i++) {
                    int tupleId = Sh.getElementOfSetS(seqElem,i);
                    if(Math.abs(h.getValue(tupleId)) > 1e-6){
                        addToH = true;
                        break;
                    }
                }
            }
            if(addToH){
                // calculate delta and extend set S
                double delta = 0;
                for (int i = 0; i < Sh.getSizeOfSetS(seqElem); i++) {
                    int tupleId = Sh.getElementOfSetS(seqElem,i);
                    S.add(tupleId); // extend set S
                    //System.out.println("Adding to S: " + tupleId);
                    if(h.getValue(tupleId) > -1){
                        delta = Math.max(delta, (-1-h.getValue(tupleId))/(Sh.findValueOfVecHForTupleIdAndOrder(seqElem,tupleId)));
                    }
                }
                // perform step
                for (int i = 0; i < Sh.getNumberOfNonzeroElementsOfVecH(seqElem); i++) {
                    int tupleId = Sh.getNonZeroTupleIdOfVecH(seqElem,i);
                    double value = Sh.getNonZeroValueOfVecH(seqElem,i);
                    h.addValue(tupleId,delta*value);
                }
            }
        }
    }

    public static void runMPLPPP(double[] weights, IndexedInstance indexedInstance){
        double bound = calculateBound(weights, indexedInstance);
        while(true){
            System.out.println("MPLP++: " + bound);
            for (int edgeId = 1; edgeId <= indexedInstance.edges(); edgeId++) {
                // update single edge
                int from = indexedInstance.edgeIdFrom(edgeId);
                int to = indexedInstance.edgeIdTo(edgeId);

                // put weights from nodes to edge
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        int tupleIdFrom = indexedInstance.nodeIdLabelToTupleId(from,label1);
                        int tupleIdTo = indexedInstance.nodeIdLabelToTupleId(to,label2);
                        weights[tupleIdForEdge] += weights[tupleIdFrom]+weights[tupleIdTo];
                    }
                }
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) weights[indexedInstance.nodeIdLabelToTupleId(from,label1)] = 0;
                for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) weights[indexedInstance.nodeIdLabelToTupleId(to,label2)] = 0;

                // perform MPLP update
                // calculate maxima over label2, multiply by 1/2
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    double maxValue = -Double.MAX_VALUE;
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleIdForEdge]);
                    }
                    weights[indexedInstance.nodeIdLabelToTupleId(from,label1)] = maxValue/2;
                }
                // calculate maxima over label1, multiply by 1/2
                for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                    double maxValue = -Double.MAX_VALUE;
                    for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleIdForEdge]);
                    }
                    weights[indexedInstance.nodeIdLabelToTupleId(to,label2)] = maxValue/2;
                }
                // subtract from edge tuples
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        int tupleIdFrom = indexedInstance.nodeIdLabelToTupleId(from,label1);
                        int tupleIdTo = indexedInstance.nodeIdLabelToTupleId(to,label2);
                        weights[tupleIdForEdge] -= weights[tupleIdFrom]+weights[tupleIdTo];
                    }
                }


                // continue with improved MPLP++ update
                // calculate maxima over label2, subtract from edge, add to node
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    double maxValue = -Double.MAX_VALUE;
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleIdForEdge]);
                    }
                    weights[indexedInstance.nodeIdLabelToTupleId(from,label1)] += maxValue; // add to node
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        weights[tupleIdForEdge] -= maxValue; // subtract from edge
                    }
                }
                // calculate maxima over label1, subtract from edge, add to node
                for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                    double maxValue = -Double.MAX_VALUE;
                    for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleIdForEdge]);
                    }
                    weights[indexedInstance.nodeIdLabelToTupleId(to,label2)] += maxValue; // add to node
                    for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                        int tupleIdForEdge = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        weights[tupleIdForEdge] -= maxValue; // subtract from edge
                    }
                }
            }

            double newBound = calculateBound(weights,indexedInstance);
            if(newBound >= bound-Math.abs((bound*1e-7))){
                break;
            }
            bound = newBound;
        }
    }

    private double calculateEpsilon(double[] weights){
        double minValue = Double.MAX_VALUE;
        double maxValue = -Double.MAX_VALUE;
        int nodeId = 1;
        for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
            int tupleId = indexedInstance.nodeIdLabelToTupleId(nodeId,label);
            maxValue = Math.max(maxValue,weights[tupleId]);
            minValue = Math.min(minValue,weights[tupleId]);
        }
        double epsilon = maxValue-minValue;
        minValue = Double.MAX_VALUE; // reinitialize
        maxValue = -Double.MAX_VALUE;
        int edgeId = 1;
        for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); label1++) {
            for (int label2 = 0; label2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); label2++) {
                int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                maxValue = Math.max(maxValue,weights[tupleId]);
                minValue = Math.min(minValue,weights[tupleId]);
            }
        }
        epsilon += maxValue-minValue;
        return epsilon;
    }

    private static double calculateBound(double[] weights, IndexedInstance indexedInstance){
        double result = 0;
        for (int scope = 1; scope <= indexedInstance.scopes(); scope++) {
            double maxValue = -Double.MAX_VALUE;
            if(indexedInstance.arity(scope) == 1){
                // node
                int nodeId = indexedInstance.scopeIdToNodeId(scope);
                for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                    int tupleId = indexedInstance.nodeIdLabelToTupleId(nodeId,label);
                    maxValue = Math.max(maxValue,weights[tupleId]);
                }
                // maxValue calculated
                result += maxValue;

            } else {
                // edge
                int edgeId = indexedInstance.scopeIdToEdgeId(scope);
                int from = indexedInstance.edgeIdFrom(edgeId);
                int to = indexedInstance.edgeIdTo(edgeId);
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleId]);
                    }
                }
                // maxValue calculated
                result += maxValue;
            }
        }
        return result;
    }

    private boolean[] calculateActiveTuples(double[] weights, double epsilon, boolean[] result){
        //boolean[] result = new boolean[weights.length];
        for (int scope = 1; scope <= indexedInstance.scopes(); scope++) {
            double maxValue = -Double.MAX_VALUE;
            if(indexedInstance.arity(scope) == 1){
                // node
                int nodeId = indexedInstance.scopeIdToNodeId(scope);
                for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                    int tupleId = indexedInstance.nodeIdLabelToTupleId(nodeId,label);
                    maxValue = Math.max(maxValue,weights[tupleId]);
                }
                // maxValue calculated
                for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                    int tupleId = indexedInstance.nodeIdLabelToTupleId(nodeId,label);
                    if(weights[tupleId] >= maxValue-epsilon){
                        result[tupleId] = true;
                    } else {
                        result[tupleId] = false;
                    }
                }

            } else {
                // edge
                int edgeId = indexedInstance.scopeIdToEdgeId(scope);
                int from = indexedInstance.edgeIdFrom(edgeId);
                int to = indexedInstance.edgeIdTo(edgeId);
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        maxValue = Math.max(maxValue,weights[tupleId]);
                    }
                }
                // maxValue calculated
                for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
                        int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        if(weights[tupleId] >= maxValue-epsilon){
                            result[tupleId] = true;
                        } else {
                            result[tupleId] = false;
                        }
                    }
                }
            }
        }
        return result;
    }
}

/*if(bound == -21){
        System.out.println("Changing tuple " + tupleId + " from " + weights[tupleId] + " to " + (weights[tupleId] + stepSize*h.getValue(tupleId)));
        int scopeId = indexedInstance.tupleIdToScopeId(tupleId);
        if(indexedInstance.arity(scopeId) == 1){
        System.out.println("  (Node: " + indexedInstance.scopeIdToNodeId(scopeId) + " label: " + indexedInstance.tupleIdToLabel1(tupleId) + "" +
        " activity: " + csp[tupleId]);
        } else {
        System.out.println("  (Edge: " + indexedInstance.scopeIdToEdgeId(scopeId) + " labels: " + indexedInstance.tupleIdToLabel1(tupleId) +
        " " + indexedInstance.tupleIdToLabel2(tupleId) + " activity: " + csp[tupleId]);
        }
        double maxValue = -Double.MAX_VALUE;
        if(indexedInstance.arity(scopeId) == 1){
        // node
        int nodeId = indexedInstance.scopeIdToNodeId(scopeId);
        for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
        maxValue = Math.max(maxValue,weights[indexedInstance.nodeIdLabelToTupleId(nodeId,label)]);
        }
        } else {
        // edge
        int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
        int from = indexedInstance.edgeIdFrom(edgeId);
        int to = indexedInstance.edgeIdTo(edgeId);
        for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
        for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
        maxValue = Math.max(maxValue, weights[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId, label1, label2)]);
        }
        }
        }
        int numActives = 0;
        if(indexedInstance.arity(scopeId) == 1){
        // node
        int nodeId = indexedInstance.scopeIdToNodeId(scopeId);
        for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
        if(csp[indexedInstance.nodeIdLabelToTupleId(nodeId,label)]){
        numActives++;
        }
        }
        } else {
        // edge
        int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
        int from = indexedInstance.edgeIdFrom(edgeId);
        int to = indexedInstance.edgeIdTo(edgeId);
        for (int label1 = 0; label1 < indexedInstance.domainSize(from); label1++) {
        for (int label2 = 0; label2 < indexedInstance.domainSize(to); label2++) {
        if(csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId, label1, label2)]){
        numActives++;
        }
        }
        }
        }
        System.out.println("   Max value within scope:  " + maxValue + " num actives: " + numActives);
        }*/

/*System.out.println("Tuples: " + indexedInstance.tuples());


        for (int i = 0; i < indexedInstance.scopes(); i++) {
            int scopeId = i+1;
            if(indexedInstance.arity(scopeId) == 1){
                // node
                int nodeId = indexedInstance.scopeIdToNodeId(scopeId);
                for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                    System.out.println("Node: nodeid="+nodeId + " scopeId=" + scopeId+ " label="+label + " tuple=" + indexedInstance.nodeIdLabelToTupleId
                            (nodeId,label));
                    int tupleId = indexedInstance.nodeIdLabelToTupleId(nodeId,label);
                    if(indexedInstance.tupleIdToScopeId(tupleId)!= scopeId){
                        System.out.println("ERR SCOPE");
                    }
                    if(indexedInstance.tupleIdToLabel1(tupleId)!=label){
                        System.out.println("ERR LABEL1");
                    }
                }
            } else {
                // edge
                int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
                for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); label2++) {
                        System.out.println("Edge: edgeid="+edgeId + " scopeId=" + scopeId + " from="+indexedInstance.edgeIdFrom(edgeId) + " to="+
                        + indexedInstance.edgeIdTo(edgeId) + " label1="+label1 + " label2="+label2 + " tuple="+
                        + indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2));
                        int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        if(indexedInstance.tupleIdToScopeId(tupleId)!= scopeId){
                            System.out.println("ERR SCOPE");
                        }
                        if(indexedInstance.tupleIdToLabel1(tupleId)!=label1){
                            System.out.println("ERR LABEL1");
                        }
                        if(indexedInstance.tupleIdToLabel2(tupleId)!=label2){
                            System.out.println("ERR LABEL2");
                        }
                    }
                }
                edgeId = -indexedInstance.scopeIdToEdgeId(scopeId);
                for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); label1++) {
                    for (int label2 = 0; label2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); label2++) {
                        System.out.println("Edge: edgeid="+edgeId + " scopeId=" + scopeId + " from="+indexedInstance.edgeIdFrom(edgeId) + " to="+
                                + indexedInstance.edgeIdTo(edgeId) + " label1="+label1 + " label2="+label2 + " tuple="+
                                + indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2));
                        int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2);
                        if(indexedInstance.tupleIdToScopeId(tupleId)!= scopeId){
                            System.out.println("ERR SCOPE");
                        }
                    }
                }
            }
        }*/
//System.out.println("Bound: "+calculateBound(indexedInstance.tupleWeight));
//System.out.println("Epsilon: " + calculateEpsilon(indexedInstance.tupleWeight));