package Propagators;

import Instance.IndexedInstance;
import Utils.ArcConsistencyStruct;
import Utils.SequenceSh;
import Utils.SetsUSf;

public class ArcConsistencyPropagator implements PropagatorInterface{

    private IndexedInstance indexedInstance;
    private ArcConsistencyStruct arcConsistencyStruct;
    private SetsUSf setsU;


    public ArcConsistencyPropagator(IndexedInstance indexedInstance){
        this.indexedInstance = indexedInstance;
        arcConsistencyStruct = new ArcConsistencyStruct(indexedInstance);
    }

    public int run(boolean[] csp, SequenceSh sequenceSh, SetsUSf setsU){
        return resetSequencesAndRunFromFullQueue(csp,sequenceSh,setsU);
    }

    private int resetSequencesAndRunFromFullQueue(boolean[] csp, SequenceSh sequenceSh, SetsUSf setsU){
        sequenceSh.eraseSequence();
        setsU.erase();
        arcConsistencyStruct.fillQueue();

        return extendSequencesAndRunFromGivenQueue(indexedInstance,csp,sequenceSh,setsU,this.arcConsistencyStruct);
    }

    static int extendSequencesAndRunFromGivenQueue(IndexedInstance indexedInstance, boolean[] csp, SequenceSh sequenceSh, SetsUSf setsU, ArcConsistencyStruct
            arcConsistencyStruct){
        while(!arcConsistencyStruct.isEmpty()){
            int edgeId = arcConsistencyStruct.getElement();
            int currentNode = indexedInstance.edgeIdFrom(edgeId);
            int otherNode = indexedInstance.edgeIdTo(edgeId);
            for (int label = 0; label < indexedInstance.domainSize(currentNode); label++) { // label
                // check support for each label and support for each outgoing edge tuple
                boolean supportEdge = false;
                for (int label2 = 0; label2 < indexedInstance.domainSize(otherNode); label2++) { // label2
                    supportEdge = supportEdge || csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label,label2)];
                }
                int tupleIdForNodeLabel = indexedInstance.nodeIdLabelToTupleId(currentNode,label);
                boolean supportNode = csp[tupleIdForNodeLabel];

                //System.out.println("Supports: " + supportEdge + " " + supportNode);

                // if they do not coincide, erase one of them, extend sequence, put other elements to queue
                if(!supportEdge && supportNode){
                    // node is active but has no support on edge -> remove node label

                    // record deactivating vector in (h, S) as an element of the sequence, set csp tuple to false
                    addElementToSh(indexedInstance,sequenceSh, currentNode, label, otherNode, edgeId, true, csp);

                    // re-fill queue
                    for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(currentNode).length; i++) {
                        arcConsistencyStruct.addElement(indexedInstance.incidentEdgesToNodeId(currentNode)[i]);
                    }

                    // check if domain-wipeout occurred
                    boolean atLeastOne = false;
                    for (int label1 = 0; label1 < indexedInstance.domainSize(currentNode); label1++) {
                        atLeastOne = atLeastOne || csp[indexedInstance.nodeIdLabelToTupleId(currentNode,label1)];
                    }
                    if(!atLeastOne){
                        // domain wipe-out occurred
                        setsU.add(indexedInstance.nodeIdToScopeId(currentNode));
                        return 0; // empty AC closure
                    }

                } else if (supportEdge && !supportNode){
                    // edge is active but has no support in node -> remove edge tuples
                    // record deactivating vector in (h, S) as an element of the sequence, set csp tuples to false
                    addElementToSh(indexedInstance, sequenceSh, currentNode, label, otherNode, edgeId, false, csp);

                    // re-fill queue
                    arcConsistencyStruct.addElement(-edgeId);

                    // check if domain-wipeout occurred
                    boolean atLeastOne = false;
                    for (int label1 = 0; label1 < indexedInstance.domainSize(currentNode); label1++) {
                        for (int label2 = 0; label2 < indexedInstance.domainSize(otherNode); label2++) {
                            atLeastOne = atLeastOne || csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2)];
                        }
                    }
                    if(!atLeastOne){
                        // domain wipe-out occurred
                        setsU.add(indexedInstance.edgeIdToScopeId(edgeId));
                        return 0; // empty AC closure
                    }
                }
            }
        }
        return 1; // non-empty AC closure
    }

    private static void addElementToSh(IndexedInstance indexedInstance, SequenceSh sequenceSh, int currentNode, int currentLabel, int otherNode, int edgeId,
                                       boolean nodeDeactivated, boolean[] csp){
        // deactivates corresponding tuple in csp, sets new element of sequence and fills in values
        // if node should be deactivated, it will have -1
        sequenceSh.startNewElementInSequence();

        int tupleIdForNodeLabel = indexedInstance.nodeIdLabelToTupleId(currentNode, currentLabel);

        double nodeSign;
        if(nodeDeactivated) {
            nodeSign = -1;
            sequenceSh.addElementToS(tupleIdForNodeLabel);
            csp[tupleIdForNodeLabel] = false;
        } else {
            nodeSign = 1;
            for (int label2 = 0; label2 < indexedInstance.domainSize(otherNode); label2++) { // label2
                int tupleEdgeId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,currentLabel,label2);
                if(csp[tupleEdgeId]) {
                    // active edge tuple
                    sequenceSh.addElementToS(tupleEdgeId);
                    csp[tupleEdgeId] = false;
                }
            }
        }

        sequenceSh.addNonZeroElementToH(tupleIdForNodeLabel, nodeSign);
        for (int label2 = 0; label2 < indexedInstance.domainSize(otherNode); label2++) { // label2
            sequenceSh.addNonZeroElementToH(indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,currentLabel,label2),-nodeSign);
        }


    }



}
