package Propagators;

import Instance.IndexedInstance;
import Utils.*;

import java.util.LinkedList;


public class CycleConsistencyPropagator implements PropagatorInterface{

    private IndexedInstance indexedInstance;
    private ArcConsistencyStruct arcConsistencyStruct;
    private SparseVector h;
    private SetOfIndices S;

    private LinkedList<int[]> cycles;

    private int[] rotatedCycle;
    private boolean[][] anchored;

    public CycleConsistencyPropagator(IndexedInstance indexedInstance) {
        this.indexedInstance = indexedInstance;
        arcConsistencyStruct = new ArcConsistencyStruct(indexedInstance);
        h = new SparseVector(indexedInstance.tuples() + 1);
        S = new SetOfIndices(indexedInstance.tuples() + 1);

        rotatedCycle = new int[indexedInstance.nodes()];
        anchored = new boolean[indexedInstance.nodes()][indexedInstance.maxDomainSize()];

        // initialize cycles
        cycles = new LinkedList<>();
        double avgDegree = indexedInstance.averageDegree();
        if (avgDegree <= 5) {
            cycles.addAll(CycleGenerator.generateCyclesOfMaxLength(indexedInstance, 4));
        } else if (avgDegree <= 10) {
            cycles.addAll(CycleGenerator.generateCyclesOfMaxLength(indexedInstance, 3));
        }
        if (cycles.isEmpty()) {
            cycles = CycleGenerator.generateFundamentalCycles(indexedInstance);
        }

        // check cycles
        /*for (int[] cycle: cycles) {
            for (int i = 0; i < cycle.length-1; i++) {
                if(indexedInstance.edgeIdTo(cycle[i])!=indexedInstance.edgeIdFrom(cycle[i+1])){
                    System.out.println("ERR: invalid cycle 1");
                }
            }
            if(indexedInstance.edgeIdTo(cycle[cycle.length-1])!=indexedInstance.edgeIdFrom(cycle[0])){
                System.out.println("ERR: invalid cycle 0");
            }
        }*/
    }

    public int run(boolean[] csp, SequenceSh sequenceSh, SetsUSf setsU){
        sequenceSh.eraseSequence();
        setsU.erase();
        arcConsistencyStruct.fillQueue();
        int checkedCyclesWithoutRemoval = 0;
        int lastTriedCycle = 0;

        // interleave running AC with SAC
        while(true){
            // enforce AC
            int resultAC = ArcConsistencyPropagator.extendSequencesAndRunFromGivenQueue(indexedInstance,csp,sequenceSh,setsU,arcConsistencyStruct);
            if(resultAC == 0){
                return 0;
            }
            if(cycles.isEmpty()) return 0;
            // try CC
            int triedCycle = lastTriedCycle;
            boolean shouldBreak = false;
            while(true){
                int[] cycle =  cycles.get(triedCycle);
                for (int orderNodeInCycle = 0; orderNodeInCycle < cycle.length; orderNodeInCycle++) {
                    int nodeId = indexedInstance.edgeIdFrom(cycle[orderNodeInCycle]);
                    for (int label = 0; label < indexedInstance.domainSize(nodeId); label++) {
                        if(csp[indexedInstance.nodeIdLabelToTupleId(nodeId,label)]){
                            // try Cycle consistency
                            if(tryCC(orderNodeInCycle, label, csp, h, S, cycle)==0){
                                checkedCyclesWithoutRemoval = 0;
                                lastTriedCycle = triedCycle;

                                // deactivate elements from set S, initialize arc consistency queue
                                for (int i = 0; i < S.getNumberOfElements(); i++) {
                                    csp[S.getElement(i)] = false;
                                    int scopeId = indexedInstance.tupleIdToScopeId(S.getElement(i));
                                    int edgeId = indexedInstance.scopeIdToEdgeId(scopeId);
                                    arcConsistencyStruct.addElement(edgeId);
                                    arcConsistencyStruct.addElement(-edgeId);

                                    // check if it was the last active element in the given scope -> if so, add to setsU and terminate
                                    if(indexedInstance.arity(scopeId) == 2){
                                        boolean someActive = false;
                                        for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeId)); label1++) {
                                            for (int label2 = 0; label2 < indexedInstance.domainSize(indexedInstance.edgeIdTo(edgeId)); label2++) {
                                                someActive = someActive || csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2)];
                                                if(someActive) break;
                                            }
                                            if(someActive) break;
                                        }
                                        if(!someActive){
                                            setsU.add(scopeId);
                                        }
                                    } else {
                                        System.out.println("ERR: Should deactivate only edge values");
                                    }
                                }

                                // add to sequence
                                sequenceSh.startNewElementInSequence();
                                for (int i = 0; i < S.getNumberOfElements(); i++) {
                                    sequenceSh.addElementToS(S.getElement(i));
                                }
                                for (int i = 0; i < h.numberOfNonzeroElements(); i++) {
                                    int index = h.getIndexOfNthNonzeroElement(i);
                                    sequenceSh.addNonZeroElementToH(index,h.getValue(index));
                                }

                                if(setsU.numElements() > 0){
                                    return 0;
                                }

                                shouldBreak = true;
                                break;
                            }
                        }
                    }
                    if(shouldBreak){
                        break;
                    }
                }
                if(shouldBreak){
                    break;
                }
                // move to next node
                checkedCyclesWithoutRemoval++;
                if(checkedCyclesWithoutRemoval >= cycles.size()+2){
                    return 1; // instance is CC
                }
                triedCycle++;
                if(triedCycle >= cycles.size()){
                    triedCycle = 0;
                }

            }
        }
    }

    private int tryCC(int orderNodeInCycle, int chosenLabel, boolean[] csp, SparseVector h, SetOfIndices S, int[] cycle){
        // tries CC on given node and label in given cycle, returns 0 if it is not CC (thus also h and S), otherwise 1
        // does not change any indices in csp
        // first, erase h and S
        h.erase();
        S.erase();

        // rotate cycle
        int lengthOfCycle = cycle.length;
        for (int i = 0; i < lengthOfCycle; i++) {
            rotatedCycle[i] = cycle[(orderNodeInCycle+i)%cycle.length];
        }
        // initialize anchors to false
        for (int i = 0; i < lengthOfCycle; i++) {
            for (int label = 0; label < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[i])); label++) {
                anchored[i][label] = false;
            }
        }

        // initialize anchors in first node
        for (int label = 0; label < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[0])); label++) {
            if(label == chosenLabel){
                anchored[0][label] = false;
            } else {
                anchored[0][label] = true;
            }
        }

        // try to deactivate outgoing links from first node
        for (int label = 0; label < indexedInstance.domainSize(indexedInstance.edgeIdTo(rotatedCycle[0])); label++) {
            int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[0],chosenLabel,label);
            if(csp[tupleId]){
                S.add(tupleId);
                h.setValue(tupleId,-1);
                anchored[1][label] = true;
            }
        }

        // identify anchored nodes
        for (int j = 1; j < lengthOfCycle-1; j++) {
            for (int labelPrev = 0; labelPrev < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[j])); labelPrev++) {
                for (int labelNext = 0; labelNext < indexedInstance.domainSize(indexedInstance.edgeIdTo(rotatedCycle[j])); labelNext++) {
                    if(anchored[j][labelPrev] && csp[indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[j],labelPrev,labelNext)]) {
                        anchored[j+1][labelNext] = true;
                    }
                }
            }
        }

        // is there path to the node itself?
        for (int labelsLastNode = 0; labelsLastNode < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[lengthOfCycle-1])); labelsLastNode++) {
            if(anchored[lengthOfCycle-1][labelsLastNode] &&
                    csp[indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[lengthOfCycle-1],labelsLastNode,chosenLabel)]){
                return 1;
            }
        }

        // no path to node itself
        // add epsilon to non-active edges
        for (int j = 1; j < lengthOfCycle-1; j++) {
            for (int labelPrev = 0; labelPrev < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[j])); labelPrev++) {
                for (int labelNext = 0; labelNext < indexedInstance.domainSize(indexedInstance.edgeIdTo(rotatedCycle[j])); labelNext++) {
                    int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[j],labelPrev,labelNext);
                    if(anchored[j][labelPrev] && !anchored[j+1][labelNext] && !csp[tupleId]) {
                        h.setValue(tupleId,1);
                    }
                }
            }
        }

        // add epsilon to last non-active edges
        for (int labelsLastNode = 0; labelsLastNode < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[lengthOfCycle-1])); labelsLastNode++) {
            int tupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[lengthOfCycle-1],labelsLastNode,chosenLabel);
            if(anchored[lengthOfCycle-1][labelsLastNode] && !anchored[0][chosenLabel] && !csp[tupleId]){
                h.setValue(tupleId,1);
            }
        }

        // check h values
        /*int[] labelling = new int[cycle.length];
        double minValue = 0;
        while(true){
            double currValue = 0;
            // evaluate given cycle in h
            for (int i = 0; i < cycle.length; i++) {
                currValue += h.getValue(indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[i],labelling[i],labelling[(i+1)%cycle.length]));
            }
            minValue = Math.min(currValue,minValue);

            if(minValue < 0){
                System.out.println("");
                for (int j = 0; j < lengthOfCycle; j++) {
                    for (int labelPrev = 0; labelPrev < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[j])); labelPrev++) {
                        for (int labelNext = 0; labelNext < indexedInstance.domainSize(indexedInstance.edgeIdTo(rotatedCycle[j])); labelNext++) {
                            System.out.println("Edge order " + j + ", edge " + rotatedCycle[j] + "(from " +
                                    indexedInstance.edgeIdFrom(rotatedCycle[j])+ " to" +
                                    indexedInstance.edgeIdTo(rotatedCycle[j])+ ") labels: " + labelPrev + " " +
                                    labelNext +
                                    " " +
                                    "activity: " +
                                    csp[indexedInstance.edgeIdLabel1Label2ToTupleId(rotatedCycle[j],labelPrev,labelNext)]);
                        }
                    }

                    for (int labelPrev = 0; labelPrev < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[j])); labelPrev++) {
                            System.out.println("Node " +
                                    indexedInstance.edgeIdFrom(rotatedCycle[j]) + " label: " + labelPrev + "activity: " +
                                    csp[indexedInstance.nodeIdLabelToTupleId(indexedInstance.edgeIdFrom(rotatedCycle[j]),labelPrev)]);
                    }
                }

                System.out.println("ERROR: negative entries, breakpoint.");
            }

            // pass to next labelling:
            int i;
            for (i = 0; i < cycle.length+1; i++) {
                if(i == cycle.length){
                    break;
                }
                if(labelling[i]+1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(rotatedCycle[i]))){
                    break;
                }
            }
            if(i== cycle.length){
                break;
            } else {
                for (int j = 0; j < i; j++) {
                    labelling[j] = 0;
                }
                labelling[i]++;
            }

        }*/


        return 0;
    }
}
