package Propagators;

import Instance.IndexedInstance;
import Utils.*;


public class SingletonArcConsistencyPropagator implements PropagatorInterface {

    private IndexedInstance indexedInstance;
    private ArcConsistencyStruct arcConsistencyStruct;
    private ArcConsistencyStruct ACStructForSAC;
    private boolean[] cspCopy;
    private SparseVector h;
    private SetOfIndices S;
    private KillerStruct killer;
    private boolean[] processed;
    private int[] degree;
    private QueueUnique queueGraph;

    public SingletonArcConsistencyPropagator(IndexedInstance indexedInstance) {
        this.indexedInstance = indexedInstance;
        arcConsistencyStruct = new ArcConsistencyStruct(indexedInstance);
        ACStructForSAC = new ArcConsistencyStruct(indexedInstance);
        cspCopy = new boolean[indexedInstance.tuples()+1];
        h = new SparseVector(indexedInstance.tuples()+1);
        S = new SetOfIndices(indexedInstance.tuples()+1);
        killer = new KillerStruct(indexedInstance.tuples()+1);
        processed = new boolean[indexedInstance.tuples()+1];
        degree = new int[indexedInstance.tuples()+1];
        queueGraph = new QueueUnique(indexedInstance.tuples()+1);
    }

    public int run(boolean[] csp, SequenceSh sequenceSh, SetsUSf setsU){
        sequenceSh.eraseSequence();
        setsU.erase();
        arcConsistencyStruct.fillQueue();
        int checkedNodesInSACWithoutRemoval = 0;
        int lastTriedNodeInSAC = 0;

        // interleave running AC with SAC
        while(true){
            // enforce AC
            int resultAC = ArcConsistencyPropagator.extendSequencesAndRunFromGivenQueue(indexedInstance,csp,sequenceSh,setsU,arcConsistencyStruct);
            if(resultAC == 0){
                return 0;
            }
            // try SAC
            int triedNodeInSAC = lastTriedNodeInSAC;
            while(true){
                // evaluate if current node is suitable for SAC testing
                int activeLabels = 0;
                for (int label = 0; label < indexedInstance.domainSize(triedNodeInSAC); label++) {
                    if(csp[indexedInstance.nodeIdLabelToTupleId(triedNodeInSAC,label)]){
                        activeLabels++;
                    }
                }
                if(activeLabels >= 2){
                    // try SAC
                    boolean shouldBreak = false;
                    for (int label = 0; label < indexedInstance.domainSize(triedNodeInSAC); label++) {
                        if(csp[indexedInstance.nodeIdLabelToTupleId(triedNodeInSAC,label)]) {
                            System.arraycopy(csp, 0, cspCopy, 0, csp.length); // copy CSP to allow changing in propagator
                            //System.out.println("Trying SAC on node " + triedNodeInSAC + " label " + label);
                            if (trySAC(triedNodeInSAC, label, cspCopy, h, S) == 0) {
                                //System.out.println( "   -> remove node");
                                // if SAC removes given label: checkedNodesInSACWithoutRemoval = 0, lastTriedNodeInSAC = triedNodeInSAC and break -> go to enforcing AC
                                checkedNodesInSACWithoutRemoval = 0;
                                lastTriedNodeInSAC = triedNodeInSAC;

                                csp[indexedInstance.nodeIdLabelToTupleId(triedNodeInSAC,label)] = false; // remove node from CSP

                                // extend AC queue by new possibly inconsistent edges
                                for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(triedNodeInSAC).length; i++) {
                                    arcConsistencyStruct.addElement(indexedInstance.incidentEdgesToNodeId(triedNodeInSAC)[i]);
                                }

                                // add to sequence
                                sequenceSh.startNewElementInSequence();
                                for (int i = 0; i < S.getNumberOfElements(); i++) {
                                    sequenceSh.addElementToS(S.getElement(i));
                                }
                                for (int i = 0; i < h.numberOfNonzeroElements(); i++) {
                                    int index = h.getIndexOfNthNonzeroElement(i);
                                    sequenceSh.addNonZeroElementToH(index,h.getValue(index));
                                }
                                shouldBreak = true;
                                break;
                            }
                            //System.out.println( "   -> is SAC");
                        }
                    }
                    if(shouldBreak) {
                        break;
                    }
                }
                // move to next node
                checkedNodesInSACWithoutRemoval++;
                if(checkedNodesInSACWithoutRemoval >= indexedInstance.nodes()+2){
                    return 1; // instance is SAC
                }
                triedNodeInSAC++;
                if(triedNodeInSAC > indexedInstance.nodes()){
                    triedNodeInSAC = 1;
                }
            }
        }
    }

    private int getDomainFrom(int nodeId, int specialNodeId, int chosenLabel){
        if(nodeId == specialNodeId){
            return chosenLabel;
        }
        return 0;
    }

    private int getDomainTo(int nodeId, int specialNodeId, int chosenLabel){
        if(nodeId == specialNodeId){
            return chosenLabel;
        }
        return indexedInstance.domainSize(nodeId)-1;
    }

    /*private List<Integer> getDomain(int nodeId, int specialNodeId, int chosenLabel){
        LinkedList<Integer> result = new LinkedList<>();
        if(nodeId == specialNodeId){
            result.add(chosenLabel);
            return result;
        } else {
            for (int i = 0; i < indexedInstance.domainSize(nodeId); i++) {
                result.add(i);
            }
            return result;
        }
    }*/

    private int trySAC(int nodeId, int chosenLabel, boolean[] csp, SparseVector h, SetOfIndices S){
        // tries SAC on given node and label, returns 0 if it is not SAC (thus also h and S), otherwise 1
        // may destroy csp
        // first, erase h and S
        h.erase();
        S.erase();
        killer.erase();

        // fill in queue for processing
        for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(nodeId).length; i++) {
            ACStructForSAC.addElement(indexedInstance.incidentEdgesToNodeId(nodeId)[i]);
            ACStructForSAC.addElement(-indexedInstance.incidentEdgesToNodeId(nodeId)[i]);
        }

        int scopeWithWipeOut = -1;

        while(!ACStructForSAC.isEmpty()){
            int edgeId = ACStructForSAC.getElement();
            int currentNode = indexedInstance.edgeIdFrom(edgeId);
            int otherNode = indexedInstance.edgeIdTo(edgeId);
            //for (int label: getDomain(currentNode,nodeId,chosenLabel)) { // label
            for (int label = getDomainFrom(currentNode,nodeId,chosenLabel);label<=getDomainTo(currentNode,nodeId,chosenLabel);label++) { // label
                // check support for each label and support for each outgoing edge tuple
                boolean supportEdge = false;
                //for (int label2:getDomain(otherNode,nodeId,chosenLabel)) { // label2
                for (int label2 = getDomainFrom(otherNode,nodeId,chosenLabel);label2<=getDomainTo(otherNode,nodeId,chosenLabel);label2++) { // label2
                    supportEdge = supportEdge || csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label,label2)];
                }
                int tupleIdForNodeLabel = indexedInstance.nodeIdLabelToTupleId(currentNode,label);
                boolean supportNode = csp[tupleIdForNodeLabel];

                //System.out.println("Supports: " + supportEdge + " " + supportNode);

                // if they do not coincide, erase one of them, extend sequence, put other elements to queue
                if(!supportEdge && supportNode){
                    // node is active but has no support on edge -> remove node label
                    csp[tupleIdForNodeLabel] = false;
                    // set killer structure
                    killer.initializeSetForId(tupleIdForNodeLabel);
                    //for (int label2:getDomain(otherNode,nodeId,chosenLabel)) { // label2
                    for (int label2=getDomainFrom(otherNode,nodeId,chosenLabel);label2<=getDomainTo(otherNode,nodeId,chosenLabel);label2++) { // label2
                        killer.addToLastInitializedSet(indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label,label2));
                    }

                    // re-fill queue
                    for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(currentNode).length; i++) {
                        ACStructForSAC.addElement(indexedInstance.incidentEdgesToNodeId(currentNode)[i]);
                    }

                    // check if domain-wipeout occurred
                    boolean atLeastOne = false;
                    //for (int label1:getDomain(currentNode,nodeId,chosenLabel)) {
                    for (int label1=getDomainFrom(currentNode,nodeId,chosenLabel);label1<=getDomainTo(currentNode,nodeId,chosenLabel);label1++) {
                        atLeastOne = atLeastOne || csp[indexedInstance.nodeIdLabelToTupleId(currentNode,label1)];
                    }
                    if(!atLeastOne){
                        // domain wipe-out occurred -> empty AC closure
                        scopeWithWipeOut = indexedInstance.nodeIdToScopeId(currentNode);
                    }
                } else if (supportEdge && !supportNode){
                    // edge is active but has no support in node -> remove edge tuples
                    //for (int label2:getDomain(otherNode,nodeId,chosenLabel)) { // label2
                    for (int label2=getDomainFrom(otherNode,nodeId,chosenLabel);label2<=getDomainTo(otherNode,nodeId,chosenLabel);label2++) { // label2
                        int edgeTupleId = indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label,label2);
                        if(csp[edgeTupleId]){
                            csp[edgeTupleId] = false;
                            killer.initializeSetForId(edgeTupleId);
                            killer.addToLastInitializedSet(tupleIdForNodeLabel);
                        }
                    }

                    // re-fill queue
                    ACStructForSAC.addElement(-edgeId);

                    // check if domain-wipeout occurred
                    boolean atLeastOne = false;
                    //for (int label1:getDomain(currentNode,nodeId,chosenLabel)) {
                    for (int label1=getDomainFrom(currentNode,nodeId,chosenLabel);label1<=getDomainTo(currentNode,nodeId,chosenLabel);label1++) {
                        //for (int label2:getDomain(otherNode,nodeId,chosenLabel)) {
                        for (int label2=getDomainFrom(otherNode,nodeId,chosenLabel);label2<=getDomainTo(otherNode,nodeId,chosenLabel);label2++) {
                            atLeastOne = atLeastOne || csp[indexedInstance.edgeIdLabel1Label2ToTupleId(edgeId,label1,label2)];
                        }
                    }
                    if(!atLeastOne){
                        // domain wipe-out occurred
                        scopeWithWipeOut = indexedInstance.edgeIdToScopeId(edgeId);
                    }
                }
                if(scopeWithWipeOut!=-1){
                    break;
                }
            }
            if(scopeWithWipeOut!=-1){
                break;
            }
        }

        if(scopeWithWipeOut==-1){
            return 1; // has non-empty AC closure
        }

        for (int i = 0; i < processed.length; i++) {
            processed[i] = false;
            degree[i] = 0;
        }

        // initialize queue by all tuples in scope where wipe-out occurred
        if(indexedInstance.arity(scopeWithWipeOut)==1){
            // node
            int nodeRemoved = indexedInstance.scopeIdToNodeId(scopeWithWipeOut);
            for (int label = 0; label < indexedInstance.domainSize(nodeRemoved); label++) {
                queueGraph.addElement(indexedInstance.nodeIdLabelToTupleId(nodeRemoved,label));
            }
        } else {
            // edge
            int edgeRemoved = indexedInstance.scopeIdToEdgeId(scopeWithWipeOut);
            for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeRemoved)); label1++) {
                for (int label2 = 0; label2 < indexedInstance.domainSize((indexedInstance.edgeIdTo(edgeRemoved))); label2++) {
                    queueGraph.addElement(indexedInstance.edgeIdLabel1Label2ToTupleId(edgeRemoved,label1,label2));
                }
            }
        }

        // initialize degrees of nodes in the DAG using killer structure and graph queue

        while(!queueGraph.isEmpty()){
            int tupleId = queueGraph.getElement();
            processed[tupleId] = true;
            for (int i = 0; i < killer.getSizeOfSetId(tupleId); i++) {
                int otherTupleId = killer.getNthElementOfSetId(tupleId,i);
                degree[otherTupleId]++;
                if(!processed[otherTupleId]){
                    queueGraph.addElement(otherTupleId);
                    processed[otherTupleId] = true;
                }
            }
        }

        // degrees of nodes are initialized -> process in topological order
        // initialize queue by all tuples in scope where wipe-out occurred
        if(indexedInstance.arity(scopeWithWipeOut)==1){
            // node
            int nodeRemoved = indexedInstance.scopeIdToNodeId(scopeWithWipeOut);
            for (int label = 0; label < indexedInstance.domainSize(nodeRemoved); label++) {
                queueGraph.addElement(indexedInstance.nodeIdLabelToTupleId(nodeRemoved,label));
            }
        } else {
            // edge
            int edgeRemoved = indexedInstance.scopeIdToEdgeId(scopeWithWipeOut);
            for (int label1 = 0; label1 < indexedInstance.domainSize(indexedInstance.edgeIdFrom(edgeRemoved)); label1++) {
                for (int label2 = 0; label2 < indexedInstance.domainSize((indexedInstance.edgeIdTo(edgeRemoved))); label2++) {
                    queueGraph.addElement(indexedInstance.edgeIdLabel1Label2ToTupleId(edgeRemoved,label1,label2));
                }
            }
        }

        while(!queueGraph.isEmpty()){
            int tupleId = queueGraph.getElement();
            if(killer.getSizeOfSetId(tupleId)==0){
                h.setValue(tupleId,1);
            }
            for (int i = 0; i < killer.getSizeOfSetId(tupleId); i++) {
                int otherTupleId = killer.getNthElementOfSetId(tupleId,i);
                degree[otherTupleId]--;
                if(degree[otherTupleId]==0){
                    queueGraph.addElement(otherTupleId);
                }
            }
        }
        h.setValue(indexedInstance.nodeIdLabelToTupleId(nodeId,chosenLabel),-1);
        S.add(indexedInstance.nodeIdLabelToTupleId(nodeId,chosenLabel));
        return 0;
    }
}
